/*

lab3.c

Dirichlet problem using 2D grid
11/98

Kyung-suk Lhee

processors are column major numbered
data in each processor is column major ordered

e.g. 3 X 2 grid

y-axis
---------------
(1) | (3) | (5)
0,1 | 1,1 | 2,1
----------------
(0) | (2) | (4)          (processor number)
0,0 | 1,0 | 2,0          (x,y coordinate)
---------------- x-axis

x coord = processor id / y dimension
y coord = processor id % y dimension

*/

#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <cs_grid.h>
#include "block.h"

#define CONST 20
#define TAG 1
#define ALL MPI_COMM_WORLD
#define ANY MPI_ANY_SOURCE
#define loc(i, j) *(bp->data + (j) + (i) * bp->jmax)  /* column major * /
#define loc(i, j) *(bp->data + (i) + (j) * bp->imax)  / * row major */
#define COL CS_GRID_COLMAJOR
#define ROW CS_GRID_ROWMAJOR


void gcf(double val, int rgb[]) {

/* cs_grid interface function */

  if ( val <= 0.0 ) {           /* 0.0 = black */
    rgb[0] = rgb[1] = rgb[2] = 0;
  } else if ( val <= 1.0 ) {    /* 1.0 = blue */
    rgb[0] = rgb[1] = 0;
    rgb[2] = 255 * val;
  } else if ( val <= 2.0 ) {    /* 2.0 = red */
    rgb[0] = 255 * (val - 1.0);
    rgb[1] = 0;
    rgb[2] = 255 * (2.0 - val);
  } else if ( val <= 3.0 ) {    /* 3.0 = yellow */
    rgb[0] = 255;
    rgb[1] = 255 * (val - 2.0);
    rgb[2] = 0;
  } else if ( val <= 4.0 ) {    /* 4.0 = white */
    rgb[0] = rgb[1] = 255;
    rgb[2] = 255 * (val - 3.0);
  } else {      /* white */
    rgb[0] = rgb[1] = rgb[2] = 255;
  }
}

Block* BlockCreate(int x, int y) {

  Block *bp;

  bp = (Block*)malloc(sizeof(Block));
  bp->imax = x;
  bp->jmax = y;
  bp->data = (double*)malloc(sizeof(double) * x * y);
  bp->lrsendbuff = (double*)malloc(sizeof(double) * y);
  bp->tbsendbuff = (double*)malloc(sizeof(double) * x);
  bp->lbuff = (double*)malloc(sizeof(double) * y);
  bp->rbuff = (double*)malloc(sizeof(double) * y);
  bp->tbuff = (double*)malloc(sizeof(double) * x);
  bp->bbuff = (double*)malloc(sizeof(double) * x);
  return bp;
}

void BlockInit(Block *bp, int tb, int te, int iprocs, int jprocs, int id) {

  int iend, jend, i, j;
  double norm;

  bp->tb = tb;
  bp->te = te;
  bp->iprocs = iprocs;
  bp->jprocs = jprocs;

  /* i, j coordinates and neighbors are column major ordered */
  bp->icoord = id / jprocs;
  bp->jcoord = id % jprocs;
  bp->lnbr = ((bp->icoord + iprocs - 1) % iprocs) *  jprocs + bp->jcoord;
  bp->rnbr = ((bp->icoord + 1) % iprocs) * jprocs + bp->jcoord;
  bp->bnbr = bp->icoord * jprocs + ((bp->jcoord + jprocs - 1) % jprocs);
  bp->tnbr = bp->icoord * jprocs + ((bp->jcoord + 1) % jprocs);

  iend = bp->imax - 1;
  jend = bp->jmax - 1;

  for (i=0; i<=iend; i++)
    for (j=0; j<=jend; j++)
      loc(i, j) = 0.0;
  norm = 0.0;

  /* initialize boudary */
  if (bp->jcoord == 0)              /* bottom row */
    for (i=0; i<=iend; i++) {
      loc(i, 0) = CONST;
      norm += CONST;
    }
  if (bp->jcoord == jprocs-1)       /* top row */
    for (i=0; i<=iend; i++) {
      loc(i, jend) = CONST;
      norm += CONST;
    }
  if (bp->icoord == 0)              /* leftmost column */
    for (j=0; j<=jend; j++) {
      loc(0, j) = CONST;
      norm += CONST;
    }
  if (bp->icoord == iprocs-1)       /* rightmost column */
    for (j=0; j<=jend; j++) {
      loc(iend, j) = CONST;
      norm += CONST;
    }
  norm -= CONST * 4;  /* compensate overlapped corners */
  bp->norm = norm / (bp->imax * bp->jmax);
}

void BlockSendL(Block *bp) {

  int j;

  for (j=0; j<bp->jmax; j++)
    bp->lrsendbuff[j] = loc(0, j);
  MPI_Send(bp->lrsendbuff, bp->jmax, MPI_DOUBLE, bp->lnbr, TAG, ALL);
}

void BlockSendR(Block *bp) {

  int j, iend;

  iend = bp->imax - 1;

  for (j=0; j<bp->jmax; j++)
    bp->lrsendbuff[j] = loc(iend, j);
  MPI_Send(bp->lrsendbuff, bp->jmax, MPI_DOUBLE, bp->rnbr, TAG, ALL);
}

void BlockSendB(Block *bp) {

  int i;

  for (i=0; i<bp->imax; i++)
    bp->tbsendbuff[i] = loc(i, 0);
  MPI_Send(bp->tbsendbuff, bp->imax, MPI_DOUBLE, bp->bnbr, TAG, ALL);
}

void BlockSendT(Block *bp) {

  int i, jend;

  jend = bp->jmax - 1;

  for (i=0; i<bp->imax; i++)
    bp->tbsendbuff[i] = loc(i, jend);
  MPI_Send(bp->tbsendbuff, bp->imax, MPI_DOUBLE, bp->tnbr, TAG, ALL);
}

void BlockRecvR(Block *bp) {

  MPI_Status status;

  MPI_Recv(bp->rbuff, bp->jmax, MPI_DOUBLE, bp->rnbr, TAG, ALL, &status);
}

void BlockRecvL(Block *bp) {

  MPI_Status status;

  MPI_Recv(bp->lbuff, bp->jmax, MPI_DOUBLE, bp->lnbr, TAG, ALL, &status);
}

void BlockRecvT(Block *bp) {

  MPI_Status status;

  MPI_Recv(bp->tbuff, bp->imax, MPI_DOUBLE, bp->tnbr, TAG, ALL, &status);
}

void BlockRecvB(Block *bp) {

  MPI_Status status;

  MPI_Recv(bp->bbuff, bp->imax, MPI_DOUBLE, bp->bnbr, TAG, ALL, &status);
}

void BlockStep(Block *bp) {

  int iend, jend, i, j;
  double u1, u2, norm;

  iend = bp->imax - 1;
  jend = bp->jmax - 1;
  norm = 0.0;
                                /* southwest corner */
  if (bp->icoord != 0 && bp->jcoord != 0) {
    u1 = loc(0, 0);
    u2 = (loc(1, 0) + loc(0, 1) + bp->lbuff[0] + bp->bbuff[0] +
          u1) / 5.0;
    loc(0, 0) = u2;
    norm += fabs(u2 - u1);
  }

  if (bp->icoord != 0)          /* leftmost column */
    for (j=1; j<jend; j++) {
      u1 = loc(0, j);
      u2 = (loc(1, j) + bp->lbuff[j] + loc(0, j+1) +
            loc(0, j-1) + u1) / 5.0;
      loc(0, j) = u2;
      norm += fabs(u2 - u1);
    }

  if (bp->jcoord != 0)          /* bottom row */
    for (i=1; i<iend; i++) {
      u1 = loc(i, 0);
      u2 = (loc(i+1, 0) + loc(i-1, 0) + bp->bbuff[i] +
            loc(i, 1) + u1) / 5.0;
      loc(i, 0) = u2;
      norm += fabs(u2 - u1);
    }

  for (j=1; j<jend; j++)         /* center */
    for (i=1; i<iend; i++) {
      u1 = loc(i, j);
      u2 = (loc(i+1, j) + loc(i-1, j) + loc(i, j-1) +
            loc(i, j+1) + u1) / 5.0;
      loc(i, j) = u2;
      norm += fabs(u2 - u1);
    }
                                /* northwest corner */
  if (bp->icoord != 0 && bp->jcoord != bp->jprocs-1) {
    u1 = loc(0, jend);
    u2 = (loc(1, jend) + loc(0, jend-1) + bp->lbuff[jend] +
          bp->tbuff[0] + u1) / 5.0;
    loc(0, jend) = u2;
    norm += fabs(u2 - u1);
  }
                                /* southeast corner */
  if (bp->icoord != bp->iprocs-1 && bp->jcoord != 0) {
    u1 = loc(iend, 0);
    u2 = (loc(iend-1, 0) + loc(iend, 1) + bp->rbuff[0] +
          bp->bbuff[iend] + u1) / 5.0;
    loc(iend, 0) = u2;
    norm += fabs(u2 - u1);
  }

  if (bp->icoord != bp->iprocs-1) /* rightmost column */
    for (j=1; j<jend; j++) {
      u1 = loc(iend, j);
      u2 = (bp->rbuff[j] + loc(iend-1, j) + loc(iend, j+1) +
            loc(iend, j-1) + u1) / 5.0;
      loc(iend, j) = u2;
      norm += fabs(u2 - u1);
    }

  if (bp->jcoord != bp->jprocs-1) /* top row */
    for (i=1; i<iend; i++) {
      u1 = loc(i, jend);
      u2 = (bp->tbuff[i] + loc(i-1, jend) + loc(i+1, jend) +
            loc(i, jend-1) + u1) / 5.0;
      loc(i, jend) = u2;
      norm += fabs(u2 - u1);
    }
                                /* northeast corner */
  if (bp->jcoord != bp->jprocs-1 && bp->icoord != bp->iprocs-1) {
    u1 = loc(iend, jend);
    u2 = (loc(iend-1, jend) + loc(iend, jend-1) + bp->tbuff[iend] +
          bp->rbuff[jend] + u1) / 5.0;
    loc(iend, jend) = u2;
    norm += fabs(u2 - u1);
  }

  bp->norm = norm / (bp->imax * bp->jmax);
  bp->tb = bp->tb + 1;
}

void BlockNorm(Block *bp) {

  MPI_Reduce(&(bp->norm), &(bp->gnorm), 1, MPI_DOUBLE, MPI_SUM, 0, ALL);
  if (bp->icoord == 0 && bp->jcoord == 0)
    printf("global norm is:%g\n", bp->gnorm / (bp->iprocs * bp->jprocs));
}

void BlockDestroy(Block *bp) {

  free(bp->data);
  free(bp->lrsendbuff);
  free(bp->tbsendbuff);
  free(bp->lbuff);
  free(bp->rbuff);
  free(bp->tbuff);
  free(bp->bbuff);
  free(bp);
}

void main(int argc, char **argv) {

  int id, size, imax, jmax, tb, te, i, j, iblk, jblk, tmp, isize, jsize, x, y;
  Block *bp;
  double *data, *data2;
  CS_Grid grid1, grid2;
  CS_Window win;
  MPI_Status status;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(ALL, &size);
  MPI_Comm_rank(ALL, &id);

  /* commans line args */
  if (argc < 7) return;
  imax = atoi(argv[1]);
  jmax = atoi(argv[2]);
  isize = atoi(argv[3]);
  jsize = atoi(argv[4]);
  tb = atoi(argv[5]);
  te = atoi(argv[6]);
  if (isize * jsize != size ||
      imax % isize != 0 || jmax % jsize != 0) {
    if (id == 0)
      printf("invalid argument. exiting...\n");
    return;
  }

  /* initialize and display initial data */
  iblk = imax / isize;
  jblk = jmax / jsize;
  bp = BlockCreate(iblk, jblk);
  BlockInit(bp, tb, te, isize, jsize, id);
  BlockNorm(bp);
  if (id == 0) {
    data = (double*)malloc(sizeof(double) * iblk * jblk);
    grid1 = cs_grid_open(imax, jmax, gcf, "initial grid");
    memcpy(data, bp->data, sizeof(double) * iblk * jblk);
    x = y = 0;
    cs_grid_showpart(grid1, x, y, iblk, jblk, iblk, jblk, 0, 0, data, COL);
    for (j=1; j<size; j++) {
      MPI_Send(&tmp, 1, MPI_INT, j, TAG, ALL);
      MPI_Recv(data, iblk*jblk, MPI_DOUBLE, j, TAG, ALL, &status);
      x = (j / jsize) * iblk;
      y = (j % jsize) * jblk;
      cs_grid_showpart(grid1, x, y, iblk, jblk, iblk, jblk, 0, 0, data, COL);
    }
  }
  else {
    MPI_Recv(&tmp, 1, MPI_INT, 0, TAG, ALL, &status); /* ANY doesn't work? */
    MPI_Send(bp->data, iblk*jblk, MPI_DOUBLE, 0, TAG, ALL);
  }

  /* difusion */
  while (bp->tb <= bp->te) {
    BlockSendL(bp);
    BlockRecvR(bp);
    BlockSendR(bp);
    BlockRecvL(bp);
    BlockSendT(bp);
    BlockRecvB(bp);
    BlockSendB(bp);
    BlockRecvT(bp);
    BlockStep(bp);
    BlockNorm(bp);
  }

  /* display result */
  if (id == 0) {
    grid2 = cs_grid_open(imax, jmax, gcf, "result grid");
    memcpy(data, bp->data, sizeof(double) * iblk * jblk);
    x = y = 0;
    cs_grid_showpart(grid2, x, y, iblk, jblk, iblk, jblk, 0, 0, data, COL);
    for (j=1; j<size; j++) {
      MPI_Send(&tmp, 1, MPI_INT, j, TAG, ALL);
      MPI_Recv(data, iblk*jblk, MPI_DOUBLE, j, TAG, ALL, &status);
      x = (j / jsize) * iblk;
      y = (j % jsize) * jblk;
      cs_grid_showpart(grid2, x, y, iblk, jblk, iblk, jblk, 0, 0, data, COL);
    }
    win = cs_make_window ("white", 200, 300, 300, 75);
    cs_set_color ("midnightblue");
    cs_text_size (30);
    cs_printf (win, 5, 25, "click here to exit");
    while (cs_mouse_btn (win) == 0) ;
    cs_remove_window(win);
    cs_grid_close(grid1);
    cs_grid_close(grid2);
  }
  else {
    MPI_Recv(&tmp, 1, MPI_INT, 0, TAG, ALL, &status);
    MPI_Send(bp->data, iblk*jblk, MPI_DOUBLE, 0, TAG, ALL);
  }
  BlockDestroy(bp);
  MPI_Finalize();
}

