/*

CSE 791 High Performance Distributed Systems
Homework #1

smooth.c
--------

image smoothing program in MPI

Kyung-Suk Lhee
klhee@syr.edu

- one master + square mesh of worker processors required
- processors are row major numbered
- data in each processor is row major ordered


e.g. 3 X 3 grid

y-axis
---------------
(1) | (2) | (3)
0,0 | 0,1 | 0,2
----------------
(4) | (5) | (6)
1,0 | 1,1 | 1,2
----------------
(7) | (8) | (9)          (processor number)
2,0 | 2,1 | 2,2          (x,y coordinate)
---------------- x-axis

x coord = (processor id - 1) / x (or y) dimension
y coord = (processor id - 1) % x (or y) dimension

*/

#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <pnm.h>
#include "block.h"

#define TAG 1
#define ALL MPI_COMM_WORLD
#define ANY MPI_ANY_SOURCE
/*#define loc(i, j) *(bp->data + (j) + (i) * bp->jmax)  / * column major */
#define loc(i, j) *(bp->data + (i) + (j) * bp->imax)  /* row major */


Block* BlockCreate(int x, int y) {

  Block *bp;

  bp = (Block*)malloc(sizeof(Block));
  bp->xblk = x;
  bp->yblk = y;
  bp->buf2 = (xel*)malloc(sizeof(xel) * x * y);
  bp->lrsendbuff = (xel*)malloc(sizeof(xel) * y);
  bp->tbsendbuff = (xel*)malloc(sizeof(xel) * x);
  bp->lbuff = (xel*)malloc(sizeof(xel) * y);
  bp->rbuff = (xel*)malloc(sizeof(xel) * y);
  bp->tbuff = (xel*)malloc(sizeof(xel) * x);
  bp->bbuff = (xel*)malloc(sizeof(xel) * x);
  return bp;
}

void BlockInit(Block *bp, int tb, int iprocs, int jprocs, int id) {

  int iend, jend, i, j;

  bp->iprocs = iprocs;
  bp->jprocs = jprocs;

  /* i, j coordinates and neighbors are column major ordered */
  bp->icoord = id / jprocs;
  bp->jcoord = id % jprocs;
  bp->lnbr = ((bp->icoord + iprocs - 1) % iprocs) *  jprocs + bp->jcoord;
  bp->rnbr = ((bp->icoord + 1) % iprocs) * jprocs + bp->jcoord;
  bp->bnbr = bp->icoord * jprocs + ((bp->jcoord + jprocs - 1) % jprocs);
  bp->tnbr = bp->icoord * jprocs + ((bp->jcoord + 1) % jprocs);

  iend = bp->imax - 1;
  jend = bp->jmax - 1;

  for (i=0; i<=iend; i++)
    for (j=0; j<=jend; j++)
      loc(i, j) = 0.0;

  /* initialize boudary */
  if (bp->jcoord == 0)              /* bottom row */
    for (i=0; i<=iend; i++) {
      loc(i, 0) = CONST;
    }
  if (bp->jcoord == jprocs-1)       /* top row */
    for (i=0; i<=iend; i++) {
      loc(i, jend) = CONST;
    }
  if (bp->icoord == 0)              /* leftmost column */
    for (j=0; j<=jend; j++) {
      loc(0, j) = CONST;
    }
  if (bp->icoord == iprocs-1)       /* rightmost column */
    for (j=0; j<=jend; j++) {
      loc(iend, j) = CONST;
    }
}

void BlockSendL(Block *bp) {

  int j;

  for (j=0; j<bp->jmax; j++)
    bp->lrsendbuff[j] = loc(0, j);
  MPI_Send(bp->lrsendbuff, bp->jmax, MPI_DOUBLE, bp->lnbr, TAG, ALL);
}

void BlockSendR(Block *bp) {

  int j, iend;

  iend = bp->imax - 1;

  for (j=0; j<bp->jmax; j++)
    bp->lrsendbuff[j] = loc(iend, j);
  MPI_Send(bp->lrsendbuff, bp->jmax, MPI_DOUBLE, bp->rnbr, TAG, ALL);
}

void BlockSendB(Block *bp) {

  int i;

  for (i=0; i<bp->imax; i++)
    bp->tbsendbuff[i] = loc(i, 0);
  MPI_Send(bp->tbsendbuff, bp->imax, MPI_DOUBLE, bp->bnbr, TAG, ALL);
}

void BlockSendT(Block *bp) {

  int i, jend;

  jend = bp->jmax - 1;

  for (i=0; i<bp->imax; i++)
    bp->tbsendbuff[i] = loc(i, jend);
  MPI_Send(bp->tbsendbuff, bp->imax, MPI_DOUBLE, bp->tnbr, TAG, ALL);
}

void BlockRecvR(Block *bp) {

  MPI_Status status;

  MPI_Recv(bp->rbuff, bp->jmax, MPI_DOUBLE, bp->rnbr, TAG, ALL, &status);
}

void BlockRecvL(Block *bp) {

  MPI_Status status;

  MPI_Recv(bp->lbuff, bp->jmax, MPI_DOUBLE, bp->lnbr, TAG, ALL, &status);
}

void BlockRecvT(Block *bp) {

  MPI_Status status;

  MPI_Recv(bp->tbuff, bp->imax, MPI_DOUBLE, bp->tnbr, TAG, ALL, &status);
}

void BlockRecvB(Block *bp) {

  MPI_Status status;

  MPI_Recv(bp->bbuff, bp->imax, MPI_DOUBLE, bp->bnbr, TAG, ALL, &status);
}

void BlockStep(Block *bp) {

  int iend, jend, i, j;
  double u1, u2;

  iend = bp->imax - 1;
  jend = bp->jmax - 1;
                                /* southwest corner */
  if (bp->icoord != 0 && bp->jcoord != 0) {
    u1 = loc(0, 0);
    u2 = (loc(1, 0) + loc(0, 1) + bp->lbuff[0] + bp->bbuff[0] +
          u1) / 5.0;
    loc(0, 0) = u2;
  }

  if (bp->icoord != 0)          /* leftmost column */
    for (j=1; j<jend; j++) {
      u1 = loc(0, j);
      u2 = (loc(1, j) + bp->lbuff[j] + loc(0, j+1) +
            loc(0, j-1) + u1) / 5.0;
      loc(0, j) = u2;
    }

  if (bp->jcoord != 0)          /* bottom row */
    for (i=1; i<iend; i++) {
      u1 = loc(i, 0);
      u2 = (loc(i+1, 0) + loc(i-1, 0) + bp->bbuff[i] +
            loc(i, 1) + u1) / 5.0;
      loc(i, 0) = u2;
    }

  for (j=1; j<jend; j++)         /* center */
    for (i=1; i<iend; i++) {
      u1 = loc(i, j);
      u2 = (loc(i+1, j) + loc(i-1, j) + loc(i, j-1) +
            loc(i, j+1) + u1) / 5.0;
      loc(i, j) = u2;
    }
                                /* northwest corner */
  if (bp->icoord != 0 && bp->jcoord != bp->jprocs-1) {
    u1 = loc(0, jend);
    u2 = (loc(1, jend) + loc(0, jend-1) + bp->lbuff[jend] +
          bp->tbuff[0] + u1) / 5.0;
    loc(0, jend) = u2;
  }
                                /* southeast corner */
  if (bp->icoord != bp->iprocs-1 && bp->jcoord != 0) {
    u1 = loc(iend, 0);
    u2 = (loc(iend-1, 0) + loc(iend, 1) + bp->rbuff[0] +
          bp->bbuff[iend] + u1) / 5.0;
    loc(iend, 0) = u2;
  }

  if (bp->icoord != bp->iprocs-1) /* rightmost column */
    for (j=1; j<jend; j++) {
      u1 = loc(iend, j);
      u2 = (bp->rbuff[j] + loc(iend-1, j) + loc(iend, j+1) +
            loc(iend, j-1) + u1) / 5.0;
      loc(iend, j) = u2;
    }

  if (bp->jcoord != bp->jprocs-1) /* top row */
    for (i=1; i<iend; i++) {
      u1 = loc(i, jend);
      u2 = (bp->tbuff[i] + loc(i-1, jend) + loc(i+1, jend) +
            loc(i, jend-1) + u1) / 5.0;
      loc(i, jend) = u2;
    }
                                /* northeast corner */
  if (bp->jcoord != bp->jprocs-1 && bp->icoord != bp->iprocs-1) {
    u1 = loc(iend, jend);
    u2 = (loc(iend-1, jend) + loc(iend, jend-1) + bp->tbuff[iend] +
          bp->rbuff[jend] + u1) / 5.0;
    loc(iend, jend) = u2;
  }
}


void BlockDestroy(Block *bp) {

  free(bp->data);
  free(bp->lrsendbuff);
  free(bp->tbsendbuff);
  free(bp->lbuff);
  free(bp->rbuff);
  free(bp->tbuff);
  free(bp->bbuff);
  free(bp);
}

void main(int argc, char **argv) {

  int id, size, bsize, i, j, k, xblk, yblk, col, row, format, num, iter[10];
  char *file[10];
  FILE *fp;
  xel **pic;
  xelval *maxval;
  Block *bp;
  MPI_Status status;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(ALL, &size);
  MPI_Comm_rank(ALL, &id);

  /* parse command line args */
  if (argc % 2 != 1) {
    if (id == 0)
      printf("invalid # of argument (should be even number). exiting...\n");
    return;
  }
  bsize = sqrt(size-1);
  if (bsize*bsize != size-1) {
    if (id == 0)
      printf("invalid # of procs (should be power of two+1). exiting...\n");
    return;
  }
  num = (argc - 1) / 2;         /* # of files */
  for (i=0; i<num; i+=2) {
    file[i] = argv[i+1];        /* file name */
    iter[i] = atoi(argv[i+2]);  /* # of iteration for the file */
  }

  /*
    master
  */

  if (id == 0) {
    for (i=0; i<num; i++) {

      /* initialize and display initial data */
      fp = fopen(file[i], "r");
      pic = pnm_readpnm(fp, &col, &row, maxval, &format);

      for (proc=1; proc<size; proc++) {

	/* 2-dim id of proc */
	x = (proc - 1) / bsize;
	y = (proc - 1) % bsize;

	/* find block size */
	xblk = col / bsize;
	yblk = row / bsize;

	if (x == bsize-1)     /* blocks on the right edge */
	  xblk = col / bsize + col % bsize;
	if (y == bsize-1)     /* blocks on the bottom edge */
	  yblk = row / bsize + row % bsize;

	/* send data size */
	MPI_Send(&xblk, 1, MPI_INT, proc, TAG, ALL);
	MPI_Send(&yblk, 1, MPI_INT, proc, TAG, ALL);

	/* send data for rendering, row by row */
	for (j=0; j<xblk; j++)
	  MPI_Send(&pic[x*xblk][y*yblk], xblk*3, MPI_CHAR, proc, TAG, ALL);
      }

      /* recv rendered data from each proc, row by row */
      for (proc=1; proc<size; proc++)
	for (j=0; j<xblk; j++)
	  MPI_Recv(&pic[x*xblk][y*yblk], xblk*3, MPI_CHAR, proc, TAG, ALL,
		   &status);
      }
  }

  /*
    worker
  */

  else {
    for (i=0; i<nfiles; i++) {

      /* recv data size */
      MPI_Recv(&xblk, 1, MPI_INT, 0, TAG, ALL, &status);
      MPI_Recv(&yblk, 1, MPI_INT, 0, TAG, ALL, &status);

      k = xblk * yblk;
      buf = (xel*)malloc(sizeof(xel) * k);

      /* recv data, row by row */
      for (j=0; j<k; j+=xblk)
	MPI_Recv(&buf[j], xblk*3, MPI_CHAR, 0, TAG, ALL, &status);

      /* create structure for worker */
      bp = BlockCreate(buf, xblk, yblk);
      BlockInit(bp, id);

      /* render, send, recv iteration */
      for (i=0; i<iter[i]; i++) {
	BlockStep(bp);
	BlockSendL(bp);
	BlockSendR(bp);
	BlockSendT(bp);
	BlockSendB(bp);
	BlockRecvL(bp);
	BlockRecvR(bp);
	BlockRecvT(bp);
	BlockRecvB(bp);
      }

      /* send rendered data to master */
      for (j=0; j<k; j+=xblk)
	MPI_Send(BlockRender(bp), xblk*3, MPI_CHAR, 0, TAG, ALL);

      BlockDestroy(bp);
    }
  }
  MPI_Finalize();
}

